# Copyright 2004-2005 Elemental Security, Inc. All Rights Reserved.
# Licensed to PSF under a Contributor Agreement.

# Modifications:
# Copyright 2006 Google, Inc. All Rights Reserved.
# Licensed to PSF under a Contributor Agreement.
# Copyright 2014 David Halter and Contributors
# Modifications are dual-licensed: MIT and PSF.

from rp.libs.parso.pgen2.generator import generate_grammar



from pyqtgraph.flowchart import Flowchart,Node
import pyqtgraph.flowchart.library as fclib
from pyqtgraph.flowchart.library.common import CtrlNode
from pyqtgraph.Qt import QtGui,QtCore
import pyqtgraph as pg
import numpy as np
import rp.r
app=QtGui.QApplication([])

## Create main window with a grid layout inside
win=QtGui.QMainWindow()
win.setWindowTitle('pyqtgraph example: FlowchartCustomNode')
cw=QtGui.QWidget()
win.setCentralWidget(cw)
layout=QtGui.QGridLayout()
cw.setLayout(layout)

## Create an empty flowchart with a single input and output
fc=Flowchart(terminals={
    'dataIn':{'io':'in'},
    'dataOut':{'io':'out'}
})
w=fc.widget()

layout.addWidget(fc.widget(),0,0,2,1)

## Create two ImageView widgets to display the raw and processed data with contrast
## and color control.
v1=pg.ImageView()
v2=pg.ImageView()
layout.addWidget(v1,0,1)
layout.addWidget(v2,1,1)

win.show()

## generate random input data
data=np.random.normal(size=(100,100))
data=25 * pg.gaussianFilter(data,(5,5))
data+=np.random.normal(size=(100,100))
data[40:60,40:60]+=15.0
data[30:50,30:50]+=15.0
# data += np.sin(np.linspace(0, 100, 1000))
# data = metaarray.MetaArray(data, info=[{'name': 'Time', 'values': np.linspace(0, 1.0, len(data))}, {}])

## Set the raw data as the input value to the flowchart
fc.setInput(dataIn=data)

## At this point, we need some custom Node classes since those provided in the library
## are not sufficient. Each node will define a set of input/output terminals, a
## processing function, and optionally a control widget (to be displayed in the
## flowchart control panel)

class ImageViewNode(Node):
    """Node that displays image data in an ImageView widget"""
    nodeName='ImageView'

    def __init__(self,name):
        self.view=None
        ## Initialize node with only a single input terminal
        Node.__init__(self,name,terminals={'data':{'io':'in'}})

    def setView(self,view):  ## setView must be called by the program
        self.view=view

    def process(self,data,display=True):
        ## if process is called with display=False, then the flowchart is being operated
        ## in batch processing mode, so we should skip displaying to improve performance.
        if display and self.view is not None:
            ## the 'data' argument is the value given to the 'data' terminal
            if data is None:
                self.view.setImage(np.zeros((1,1)))  # give a blank array to clear the view
            else:
                self.view.setImage(data)

## We will define an unsharp masking filter node as a subclass of CtrlNode.
## CtrlNode is just a convenience class that automatically creates its
## control widget based on a simple data structure.
class UnsharpMaskNode(CtrlNode):
    """Return the input data passed through an unsharp mask."""
    nodeName="UnsharpMask"
    uiTemplate=[
        ('sigma','spin',{'value':1.0,'step':1.0,'bounds':[0.0,None]}),
        ('strength','spin',{'value':1.0,'dec':True,'step':0.5,'minStep':0.01,'bounds':[0.0,None]}),
    ]
    def __init__(self,name):
        ## Define the input / output terminals available on this node
        terminals={
            'dataIn':dict(io='in'),  # each terminal needs at least a name and
            'dataOut':dict(io='out'),  # to specify whether it is input or output
        }  # other more advanced options are available
        # as well..
        CtrlNode.__init__(self,name,terminals=terminals)

    def process(self,dataIn,display=True):
        # CtrlNode has created self.ctrls, which is a dict containing {ctrlName: widget}
        sigma=self.ctrls['sigma'].value()
        strength=self.ctrls['strength'].value()
        output=dataIn - (strength * pg.gaussianFilter(dataIn,(sigma,sigma)))
        return {'dataOut':output}

## To make our custom node classes available in the flowchart context menu,
## we can either register them with the default node library or make a
## new library.


## Method 1: Register to global default library:
# fclib.registerNodeType(ImageViewNode, [('Display',)])
# fclib.registerNodeType(UnsharpMaskNode, [('Image',)])

## Method 2: If we want to make our custom node available only to this flowchart,
## then instead of registering the node type globally, we can create a new
## NodeLibrary:
library=fclib.LIBRARY.copy()  # start with the default node set
library.addNodeType(ImageViewNode,[('Display',)])
# Add the unsharp mask node to two locations in the menu to demonstrate
# that we can create arbitrary menu structures
library.addNodeType(UnsharpMaskNode,[('Image',),
                                     ('Submenu_test','submenu2','submenu3')])
fc.setLibrary(library)

## Now we will programmatically add nodes to define the function of the flowchart.
## Normally, the user will do this manually or by loading a pre-generated
## flowchart file.

v1Node=fc.createNode('ImageView',pos=(0,-150))
v1Node.setView(v1)

v2Node=fc.createNode('ImageView',pos=(150,-150))
v2Node.setView(v2)

fNode=fc.createNode('UnsharpMask',pos=(0,0))
fc.connectTerminals(fc['dataIn'],fNode['dataIn'])
fc.connectTerminals(fc['dataIn'],v1Node['data'])
fc.connectTerminals(fNode['dataOut'],v2Node['data'])
fc.connectTerminals(fNode['dataOut'],fc['dataOut'])

## Start Qt event loop unless running in interactive mode or using pyside.
if __name__ == '__main__':
    import sys
    if (sys.flags.interactive != 1) or not hasattr(QtCore,'PYQT_VERSION'):
        QtGui.QApplication.instance().exec_()


from typing import Any
from .constants import FIELD_TYPE as FIELD_TYPE, FLAG as FLAG
from .charset import charset_by_id as charset_by_id

PYTHON3: Any
ESCAPE_REGEX: Any
ESCAPE_MAP: Any

def escape_item(val, charset): ...
def escape_dict(val, charset): ...
def escape_sequence(val, charset): ...
def escape_set(val, charset): ...
def escape_bool(value): ...
def escape_object(value): ...

escape_int: Any

escape_long: Any

def escape_float(value): ...
def escape_string(value): ...
def escape_unicode(value): ...
def escape_None(value): ...
def escape_timedelta(obj): ...
def escape_time(obj): ...
def escape_datetime(obj): ...
def escape_date(obj): ...
def escape_struct_time(obj): ...
def convert_datetime(connection, field, obj): ...
def convert_timedelta(connection, field, obj): ...
def convert_time(connection, field, obj): ...
def convert_date(connection, field, obj): ...
def convert_mysql_timestamp(connection, field, timestamp): ...
def convert_set(s): ...
def convert_bit(connection, field, b): ...
def convert_characters(connection, field, data): ...
def convert_int(connection, field, data): ...
def convert_long(connection, field, data): ...
def convert_float(connection, field, data): ...

encoders: Any
decoders: Any
conversions: Any

def convert_decimal(connection, field, data): ...
def escape_decimal(obj): ...



from typing import Any, Optional

UTF8_COOKIE: Any
system_exceptions: Any
HEADER: Any
FOOTER: Any
PAGE_HTML: Any
CONSOLE_HTML: Any
SUMMARY_HTML: Any
FRAME_HTML: Any
SOURCE_LINE_HTML: Any

def render_console_html(secret, evalex_trusted: bool = ...): ...
def get_current_traceback(ignore_system_exceptions: bool = ..., show_hidden_frames: bool = ..., skip: int = ...): ...

class Line:
    lineno: Any
    code: Any
    in_frame: Any
    current: Any
    def __init__(self, lineno, code): ...
    def classes(self): ...
    def render(self): ...

class Traceback:
    exc_type: Any
    exc_value: Any
    exception_type: Any
    frames: Any
    def __init__(self, exc_type, exc_value, tb): ...
    def filter_hidden_frames(self): ...
    def is_syntax_error(self): ...
    def exception(self): ...
    def log(self, logfile: Optional[Any] = ...): ...
    def paste(self): ...
    def render_summary(self, include_title: bool = ...): ...
    def render_full(self, evalex: bool = ..., secret: Optional[Any] = ..., evalex_trusted: bool = ...): ...
    def generate_plaintext_traceback(self): ...
    def plaintext(self): ...
    id: Any

class Frame:
    lineno: Any
    function_name: Any
    locals: Any
    globals: Any
    filename: Any
    module: Any
    loader: Any
    code: Any
    hide: Any
    info: Any
    def __init__(self, exc_type, exc_value, tb): ...
    def render(self): ...
    def render_line_context(self): ...
    def get_annotated_lines(self): ...
    def eval(self, code, mode: str = ...): ...
    def sourcelines(self): ...
    def get_context_lines(self, context: int = ...): ...
    @property
    def current_line(self): ...
    def console(self): ...
    id: Any



from copy import copy, deepcopy
import rp
import time
import abc

from pynput import mouse, keyboard
from pynput.mouse import Controller as MouseController, Button
from pynput.keyboard import Controller as KeyboardController, Key

##############################################################################
# A C T I O N   C L A S S E S
##############################################################################

mouse_controller = MouseController()
keyboard_controller = KeyboardController()


class Reprable:
    def __repr__(self):
        attributes = [
            f"{attr}={str(value)}"
            for attr, value in self.__dict__.items()
            if not "__" in attr
        ]
        return f'{type(self).__name__}({", ".join(attributes)})'


class Action(abc.ABC, Reprable):
    """
    Base class for all actions. Each action is callable.
    """

    @abc.abstractmethod
    def __call__(self):
        """
        Perform the action, e.g. press a key, move the mouse, etc.
        """
        pass


class DelayAction(Action):
    """
    Sleeps for `seconds` when called.
    """

    def __init__(self, seconds: float):
        self.seconds = seconds

    def __call__(self):
        if self.seconds > 0:
            time.sleep(self.seconds)


# ------------------- Mouse Actions -------------------
class MouseAction(Action):
    """
    Base for mouse actions: provides x, y coordinates.
    """

    def __init__(self, x, y):
        self.x = x
        self.y = y


class MouseMoveAction(MouseAction):
    """
    Moves the mouse pointer to (x, y).
    """

    def __init__(self, x, y):
        super().__init__(x, y)

    def __call__(self):
        mouse_controller.position = (self.x, self.y)


class MousePressAction(MouseAction):
    """
    Presses a mouse button at (x, y).
    """

    def __init__(self, x, y, button=Button.left):
        super().__init__(x, y)
        self.button = button

    def __call__(self):
        mouse_controller.position = (self.x, self.y)
        mouse_controller.press(self.button)


class MouseReleaseAction(MouseAction):
    """
    Releases a mouse button at (x, y).
    """

    def __init__(self, x, y, button=Button.left):
        super().__init__(x, y)
        self.button = button

    def __call__(self):
        mouse_controller.position = (self.x, self.y)
        mouse_controller.release(self.button)


class MouseScrollAction(MouseAction):
    """
    Scrolls the mouse wheel at (x, y) by (dx, dy).
    """

    def __init__(self, x, y, dx, dy):
        super().__init__(x, y)
        self.dx = dx
        self.dy = dy

    def __call__(self):
        mouse_controller.position = (self.x, self.y)
        mouse_controller.scroll(self.dx, self.dy)


# ------------------- Keyboard Actions -------------------
class KeyboardAction(Action):
    """
    Base for keyboard actions. Stores a Key or a character.
    """

    def __init__(self, key):
        self.key = key


class KeyboardPressAction(KeyboardAction):
    """
    Presses a key (character or special Key).
    """

    def __init__(self, key):
        super().__init__(key)

    def __call__(self):
        keyboard_controller.press(self.key)


class KeyboardReleaseAction(KeyboardAction):
    """
    Releases a key (character or special Key).
    """

    def __init__(self, key):
        super().__init__(key)

    def __call__(self):
        keyboard_controller.release(self.key)


##############################################################################
# T R I G G E R S
##############################################################################


class Trigger(abc.ABC, Reprable):
    """
    Each trigger is a callable that takes an Action object.
    It returns True if triggered, otherwise False.
    Triggers can manage internal state as needed.
    """

    NUM_ACTIONS = 0

    @abc.abstractmethod
    def __call__(self, action: Action) -> bool:
        pass


class NoStartTrigger(Trigger):
    """
    Always returns True immediately -> recording starts right away.
    """

    def __call__(self, action: Action) -> bool:
        return True


class NoEndTrigger(Trigger):
    """
    Never returns True -> must stop via Ctrl+C or other means.
    """

    def __call__(self, action: Action) -> bool:
        return False


class KeyDownTrigger(Trigger):
    NUM_ACTIONS = 1

    def __init__(self, target_key):
        self.target_key = target_key

    def __call__(self, action: Action) -> bool:
        return (
            isinstance(action, KeyboardReleaseAction) and action.key == self.target_key
        )


class KeyUpTrigger(Trigger):
    NUM_ACTIONS = 1

    def __init__(self, target_key):
        self.target_key = target_key

    def __call__(self, action: Action) -> bool:
        return isinstance(action, KeyboardPressAction) and action.key == self.target_key


class TwoKeyDownTrigger(Trigger):
    NUM_ACTIONS = 2

    def __init__(self, key1, key2):
        self.key1 = key1
        self.key2 = key2
        self.key1_pressed = False
        self.key2_pressed = False
        self.triggered = False

    def __call__(self, action: Action) -> bool:
        if self.triggered:
            return False

        if isinstance(action, KeyboardPressAction):
            if action.key in self.key1:
                self.key1_pressed = True
            elif action.key in self.key2:
                self.key2_pressed = True
            else:
                self.key1_pressed = False
                self.key2_pressed = False

            if self.key1_pressed and self.key2_pressed:
                self.triggered = True
                return True

        elif isinstance(action, KeyboardReleaseAction):
            self.key1_pressed = False
            self.key2_pressed = False

        return False


class TwoKeyUpTrigger(Trigger):
    NUM_ACTIONS = 2

    def __init__(self, key1, key2):
        self.key1 = key1
        self.key2 = key2
        self.key1_released = False
        self.key2_released = False
        self.down_trigger = TwoKeyDownTrigger(self.key1, self.key2)
        self.triggered = False
        self.ready = False

    def __call__(self, action: Action) -> bool:
        if self.triggered:
            return False

        self.ready |= self.down_trigger(action)

        if self.ready and isinstance(action, KeyboardReleaseAction):
            self.key1_released |= action.key in self.key1
            self.key2_released |= action.key in self.key2

            if self.ready and self.key1_released and self.key2_released:
                self.triggered = True
                return True

        #FOR DEBUGGING
        # print(
        #     line_join(
        #         [
        #             fansi(self, "white green"),
        #             fansi(action, "yellow white"),
        #             fansi(self.down_trigger, "orange"),
        #         ]
        #     )
        # )

        return False


class CmdEscUpTrigger(TwoKeyUpTrigger):
    def __init__(self):
        super().__init__([Key.cmd, Key.cmd_r], [Key.esc])


class CmdEscDownTrigger(TwoKeyDownTrigger):
    def __init__(self):
        super().__init__([Key.cmd, Key.cmd_r], [Key.esc])


class CmdShiftUpTrigger(TwoKeyUpTrigger):
    def __init__(self):
        super().__init__([Key.cmd, Key.cmd_r], [Key.shift])


class CmdShiftDownTrigger(TwoKeyDownTrigger):
    def __init__(self):
        super().__init__([Key.cmd, Key.cmd_r], [Key.shift])


class CmdCmdUpTrigger(TwoKeyUpTrigger):
    # This is buggy because pynput is buggy with respect to two cmd keys! Don't use
    def __init__(self):
        super().__init__([Key.cmd], [Key.cmd_r])


class CmdCmdDownTrigger(TwoKeyDownTrigger):
    def __init__(self):
        super().__init__([Key.cmd], [Key.cmd_r])


##############################################################################
# R E C O R D I N G  W I T H  S E P A R A T E  D E L A Y S
##############################################################################


def record_actions(start_trigger=None, end_trigger=None):
    """
    Capture a list of Action objects:
      - DelayAction(seconds)
      - MouseMoveAction, MousePressAction, etc.
      - KeyboardPressAction, KeyboardReleaseAction
    The action that triggers start or end is NOT recorded.
    """
    if start_trigger is None:
        start_trigger = NoStartTrigger()
    if end_trigger is None:
        end_trigger = NoEndTrigger()

    actions = []
    start_recording = False
    stop_recording = False
    last_time = None

    def create_action(event_type, data):
        """
        Convert a (event_type, data) from pynput into one of our Action objects.
        """
        if event_type == "mouse_move":
            x, y = data
            return MouseMoveAction(x, y)

        elif event_type == "mouse_press":
            btn, (x, y) = data
            button = Button.left if str(btn) == "Button.left" else Button.right
            return MousePressAction(x, y, button)

        elif event_type == "mouse_release":
            btn, (x, y) = data
            button = Button.left if str(btn) == "Button.left" else Button.right
            return MouseReleaseAction(x, y, button)

        elif event_type == "mouse_scroll":
            x, y, dx, dy = data
            return MouseScrollAction(x, y, dx, dy)

        elif event_type == "keyboard_press":
            return KeyboardPressAction(data)

        elif event_type == "keyboard_release":
            return KeyboardReleaseAction(data)

        return None  # Unknown event

    def handle_event(event_type, data):
        nonlocal stop_recording

        if stop_recording:
            return

        nonlocal start_recording, last_time

        # Convert to an Action object
        action = create_action(event_type, data)
        if action is None:
            return  # skip unknown events

        # Check triggers

        if not start_recording:
            rp.fansi_print(action, "italic dark gray")
            if start_trigger(action):
                # Just triggered the start
                start_recording = True
                last_time = time.time()
            return  # Don't add actions before starting!
        else:
            if end_trigger(action):
                # Just triggered the end
                print("BLOOBAH")
                for _ in range(end_trigger.NUM_ACTIONS):
                    while actions and isinstance(actions[-1], DelayAction):
                        actions.pop()
                    if actions:
                        actions.pop()
                        deleted_action = actions.pop()
                        # This is ugly logic...actions should instead be FILTERED through triggers.
                        # For this reason, I'll abstract this whole codebase away into an rp function, as I might improve this logic in the future to handle new types of triggers.
                        rp.fansi_print(
                            f"[{len(actions)}]: ACTIONS -= {action}", "orange red blue green"
                        )
                stop_recording = True
                return

        # If we are here, we are actively recording this action
        now = time.time()
        if last_time is None:
            last_time = now
        delta = now - last_time
        last_time = now

        # Add a DelayAction if delta > 0
        if delta > 0:
            actions.append(DelayAction(delta))

        # Add the actual action
        actions.append(action)
        rp.fansi_print(f"[{len(actions)}]: ACTIONS += {action}", "yellow blue green")

    # ---------------------------------------------------------------------
    #  Listener callbacks
    # ---------------------------------------------------------------------
    def on_move(x, y):
        handle_event("mouse_move", (x, y))

    def on_click(x, y, button, pressed):
        if pressed:
            handle_event("mouse_press", (button, (x, y)))
        else:
            handle_event("mouse_release", (button, (x, y)))

    def on_scroll(x, y, dx, dy):
        handle_event("mouse_scroll", (x, y, dx, dy))

    def on_press(key):
        handle_event("keyboard_press", key)

    def on_release(key):
        handle_event("keyboard_release", key)

    # Start listeners
    mouse_listener = mouse.Listener(
        on_move=on_move, on_click=on_click, on_scroll=on_scroll
    )
    keyboard_listener = keyboard.Listener(on_press=on_press, on_release=on_release)
    mouse_listener.start()
    keyboard_listener.start()

    print("Recording...Press cmd+shift anywhere or Ctrl+C in the terminal to stop.")
    try:
        while not stop_recording:
            time.sleep(0.000001)
    except KeyboardInterrupt:
        print("Stopped by KeyboardInterrupt.")
    finally:
        mouse_listener.stop()
        keyboard_listener.stop()

    print(f"Recorded {len(actions)} actions (including DelayActions).")
    return actions


##############################################################################
# P L A Y B A C K
##############################################################################

def playback_actions(actions):
    """
    Replays each Action in sequence by simply calling it.
    Press escape to interrupt playback!
    """

    #A safekey
    abort=False
    abort_key=Key.esc
    def on_press(key):
        nonlocal abort
        if key==abort_key:
            abort=True
            safekey_listener.stop()
    safekey_listener = keyboard.Listener(on_press=on_press)
    safekey_listener.start()

    rp.fansi_print(f"Playing back {len(actions)} actions...press {abort_key} to abort!",'white white bold')
    for i, action in enumerate(actions, start=1):
        if abort:
            rp.fansi_print('Aborting playback!','red orange','bold')
            raise KeyboardInterrupt
        action()
        style = None
        if isinstance(action, DelayAction): style = 'dark gray italic'
        if isinstance(action, MouseMoveAction): style = 'dark light gray'
        elif isinstance(action, MouseAction): style = 'gray green yellow bold'
        if isinstance(action, KeyboardReleaseAction): style = 'white blue red blue magenta bold'
        if isinstance(action, KeyboardPressAction): style = 'white blue green blue cyan bold'
        action_str = rp.fansi(action, style)
        num_str = rp.fansi(f"[{i}/{len(actions)}]", "yellow gray")
        print(f"{num_str} {action_str}")

    rp.fansi_print("Playback finished.",'white white bold')


##############################################################################
# D E L A Y   U T I L I T I E S
##############################################################################


def lstrip_delays(actions):
    """
    If the first action is a DelayAction, set its .seconds = 0.
    """
    while actions and isinstance(actions[0], DelayAction):
        actions = actions[1:]
    return actions


def rstrip_delays(actions):
    """
    If the last action is a DelayAction, set its .seconds = 0.
    """
    while actions and isinstance(actions[-1], DelayAction):
        actions = actions[:-1]
    return actions


def strip_delays(actions):
    """
    Set the first and last DelayAction (if any) to 0.0
    """
    actions = lstrip_delays(actions)
    actions = rstrip_delays(actions)
    return actions


##############################################################################
# Default Recorder
##############################################################################


class PynputCasette(list):
    def __init__(self, actions=None):
        super().__init__([] if actions is None else actions)
        self.start_trigger = CmdShiftUpTrigger()
        self.end_trigger = CmdShiftDownTrigger()

    def record(self):
        return PynputCasette(
            self + record_actions(self.start_trigger, self.end_trigger)
        )

    def strip(self):
        return PynputCasette(strip_delays(copy(self)))

    def lstrip(self):
        return PynputCasette(lstrip_delays(copy(self)))

    def rstrip(self):
        return PynputCasette(rstrip_delays(copy(self)))

    def speedup(self, factor):
        self = deepcopy(self)
        for x in self:
            if isinstance(x, DelayAction):
                x.seconds /= factor
        return self

    def slowdown(self, factor):
        return self.speedup(1 / factor)

    def remove_mousemoves(self):
        return PynputCasette(x for x in self if not isinstance(x, MouseMoveAction)).merge_delays()

    def merge_delays(self):
        out = []
        for x in self:
            x = copy(x)
            if isinstance(x, DelayAction) and out and isinstance(out[-1], DelayAction):
                out[-1].seconds+=x.seconds
            else:
                out.append(x)
        return PynputCasette(out)

    def set_all_delays(self,func_or_float):
        """ Takes a float -> float func OR a float"""
        if callable(func_or_float):
            func = func_or_float
        else:
            func_or_float = float(func_or_float)
            func = lambda x: func_or_float

        out = []
        for x in self:
            x=copy(x)
            if isinstance(x, DelayAction):
                x.seconds=func(x.seconds)
            out.append(x)
        return PynputCasette(out)

    def cap_all_delays(self,seconds):
        return self.set_all_delays(lambda x:min(x,seconds))

    def play(self, loop=False):
        for _ in range(1 + 999**999 * loop):
            playback_actions(self)

    def repr(self):
        return f"{type(self).__name__}({repr(self.actions)}"




import urwid


class HelpBar(urwid.WidgetWrap):
    HELP_TEXT = [
        ('title', "Pyfx"), "    ", "UP, DOWN, ENTER, Q",
    ]

    def __init__(self, manager):
        self._manager = manager
        self._text_widget = urwid.Text(HelpBar.HELP_TEXT)
        super().__init__(urwid.AttrMap(self._text_widget, "foot"))



from typing import Optional

class sha224(object):
    name: str
    block_size: int
    digest_size: int
    digestsize: int
    def __init__(self, init: Optional[str]) -> None: ...
    def copy(self) -> sha224: ...
    def digest(self) -> str: ...
    def hexdigest(self) -> str: ...
    def update(self, arg: str) -> None: ...

class sha256(object):
    name: str
    block_size: int
    digest_size: int
    digestsize: int
    def __init__(self, init: Optional[str]) -> None: ...
    def copy(self) -> sha256: ...
    def digest(self) -> str: ...
    def hexdigest(self) -> str: ...
    def update(self, arg: str) -> None: ...



"""
Creation of the `Layout` instance for the Python input/REPL.
"""
from __future__ import unicode_literals

from rp.prompt_toolkit.enums import DEFAULT_BUFFER, SEARCH_BUFFER
from rp.prompt_toolkit.filters import IsDone, HasCompletions, RendererHeightIsKnown, HasFocus, Condition
from rp.prompt_toolkit.key_binding.vi_state import InputMode
from rp.prompt_toolkit.layout.containers import Window, HSplit, VSplit, FloatContainer, Float, ConditionalContainer, ScrollOffsets
from rp.prompt_toolkit.layout.controls import BufferControl, TokenListControl, FillControl, UIContent
from rp.prompt_toolkit.layout.dimension import LayoutDimension
from rp.prompt_toolkit.layout.lexers import SimpleLexer
from rp.prompt_toolkit.layout.margins import PromptMargin
from rp.prompt_toolkit.layout.menus import CompletionsMenu, MultiColumnCompletionsMenu
from rp.prompt_toolkit.layout.processors import (
    ConditionalProcessor, AppendAutoSuggestion, HighlightSearchProcessor, 
    HighlightSelectionProcessor, HighlightMatchingBracketProcessor, 
    Processor, Transformation, IndentGuideProcessor, ShowWhitespaceProcessor,
    HighlightWordOccurrencesProcessor
)
from rp.prompt_toolkit.layout.screen import Char
from rp.prompt_toolkit.layout.toolbars import CompletionsToolbar, ArgToolbar, SearchToolbar, ValidationToolbar, SystemToolbar, TokenListToolbar
from rp.prompt_toolkit.layout.utils import token_list_width
from rp.prompt_toolkit.reactive import Integer
from rp.prompt_toolkit.selection import SelectionType
from rp.prompt_toolkit.token import Token
from rp.prompt_toolkit.document import Document
from rp.prompt_toolkit.selection import SelectionState

from .filters import HasSignature, ShowSidebar, ShowSignature, ShowDocstring,ShowRealtimeInput,ShowVarSpace,ShowVarSpaceAndShowRealtimeInput,ShowVarSpaceOrShowRealtimeInput,ShowParenthesisAutomator
from .utils import if_mousedown

from pygments.lexers import PythonLexer
from pygments.token import Token

import platform
import sys

__all__ = (
    'create_layout',
    'CompletionVisualisation',
)


# DisplayMultipleCursors: Only for rp.prompt_toolkit>=1.0.8
try:
    from rp.prompt_toolkit.layout.processors import DisplayMultipleCursors
except ImportError:
    class DisplayMultipleCursors(Processor):
        " Dummy. "
        def __init__(self, *a):
            pass

        def apply_transformation(self, cli, document, lineno,
                                 source_to_display, tokens):
            return Transformation(tokens)


class CompletionVisualisation:
    " Visualisation method for the completions. "
    NONE = 'none'
    POP_UP = 'pop-up'
    MULTI_COLUMN = 'multi-column'
    TOOLBAR = 'toolbar'


def show_completions_toolbar(python_input):
    return Condition(lambda cli: python_input.completion_visualisation == CompletionVisualisation.TOOLBAR)


def show_completions_menu(python_input):
    return Condition(lambda cli: python_input.completion_visualisation == CompletionVisualisation.POP_UP)


def show_multi_column_completions_menu(python_input):
    return Condition(lambda cli: python_input.completion_visualisation == CompletionVisualisation.MULTI_COLUMN)


def python_sidebar(python_input):
    """
    Create the `Layout` for the sidebar with the configurable options.
    """
    def get_tokens(cli):
        tokens = []
        T = Token.Sidebar

        def append_category(category):
            tokens.extend([
                (T, '  '),
                (T.Title, '   %-36s' % category.title),
                (T, '\n'),
            ])

        def append(index, label, status):
            selected = index == python_input.selected_option_index

            @if_mousedown
            def select_item(cli, mouse_event):
                python_input.selected_option_index = index

            @if_mousedown
            def goto_next(cli, mouse_event):
                " Select item and go to next value. "
                python_input.selected_option_index = index
                option = python_input.selected_option
                option.activate_next()

            token = T.Selected if selected else T

            tokens.append((T, ' >' if selected else '  '))
            tokens.append((token.Label, '%-24s' % label, select_item))
            tokens.append((token.Status, ' ', select_item))
            tokens.append((token.Status, '%s' % status, goto_next))

            if selected:
                tokens.append((Token.SetCursorPosition, ''))

            tokens.append((token.Status, ' ' * (13 - len(status)), goto_next))
            tokens.append((T, '<' if selected else ''))
            tokens.append((T, '\n'))

        i = 0
        for category in python_input.options:
            append_category(category)

            for option in category.options:
                append(i, option.title, '%s' % option.get_current_value())
                i += 1

        tokens.pop()  # Remove last newline.

        return tokens

    class Control(TokenListControl):
        def move_cursor_down(self, cli):
            python_input.selected_option_index += 1

        def move_cursor_up(self, cli):
            python_input.selected_option_index -= 1

    return ConditionalContainer(
        content=Window(
            Control(get_tokens, Char(token=Token.Sidebar),
                has_focus=ShowSidebar(python_input) & ~IsDone()),
            width=LayoutDimension.exact(43),
            height=LayoutDimension(min=3),
            scroll_offsets=ScrollOffsets(top=1, bottom=1)),
        filter=ShowSidebar(python_input) & ~IsDone())


def python_sidebar_navigation(python_input):
    """
    Create the `Layout` showing the navigation information for the sidebar.
    """
    def get_tokens(cli):
        tokens = []
        T = Token.Sidebar

        # Show navigation info.
        tokens.extend([
            (T.Separator, ' ' * 43 + '\n'),
            (T, '    '),
            (T.Key, '[Arrows]'),
            (T, ' '),
            (T.Key.Description, 'Navigate'),
            (T, ' '),
            (T.Key, '[F5]'),
            (T, ' '),
            (T.Key.Description, 'Save Settings'),
        ])

        return tokens

    return ConditionalContainer(
        content=Window(
            TokenListControl(get_tokens, Char(token=Token.Sidebar)),
            width=LayoutDimension.exact(43),
            height=LayoutDimension.exact(2)),
        filter=ShowSidebar(python_input) & ~IsDone())


def python_sidebar_help(python_input):
    """
    Create the `Layout` for the help text for the current item in the sidebar.
    """
    token = Token.Sidebar.HelpText

    def get_current_description():
        """
        Return the description of the selected option.
        """
        i = 0
        for category in python_input.options:
            for option in category.options:
                if i == python_input.selected_option_index:
                    return option.description
                i += 1
        return ''

    def get_tokens(cli):
        return [(token, get_current_description())]

    return ConditionalContainer(
        content=Window(
            TokenListControl(get_tokens, Char(token=token)),
            height=LayoutDimension(min=3)),
        filter=ShowSidebar(python_input) &
               Condition(lambda cli: python_input.show_sidebar_help) & ~IsDone())


def signature_toolbar(python_input):
    """
    Return the `Layout` for the signature.
    """
    def get_tokens(cli):
        result = []
        append = result.append
        Signature = Token.Toolbar.Signature

        if python_input.signatures:
            sig = python_input.signatures[0]  # Always take the first one.

            append((Signature, ' '))
            try:
                append((Signature, sig.full_name))
            except:# IndexError but Also AttributeError: 'NoneType' object has no attribute 'split'
                # Workaround for #37: https://github.com/jonathanslenders/python-prompt-toolkit/issues/37
                # See also: https://github.com/davidhalter/jedi/issues/490
                return []

            append((Signature.Operator, '('))

            try:
                enumerated_params = enumerate(sig.params)
            except AttributeError:
                # Workaround for #136: https://github.com/jonathanslenders/ptpython/issues/136
                # AttributeError: 'Lambda' object has no attribute 'get_subscope_by_name'
                return []

            for i, p in enumerated_params:
                try:
                    # Workaround for #47: 'p' is None when we hit the '*' in the signature.
                    #                     and sig has no 'index' attribute.
                    # See: https://github.com/jonathanslenders/ptpython/issues/47
                    #      https://github.com/davidhalter/jedi/issues/598
                    description = (p.description if p else '*') #or '*'
                    sig_index = getattr(sig, 'index', 0)

                    #region Ryan Burgert Code (it was annoying seeing f(param x,param y) etc. So I got rid of the 'param' stuff and now it looks nicer
                    desc_string=str(description)
                    if desc_string.startswith("param "):
                        desc_string=desc_string[6:]
                    #endregion
                    if i == sig_index:
                        # Note: we use `_Param.description` instead of
                        #       `_Param.name`, that way we also get the '*' before args.
                        append((Signature.CurrentName,desc_string))
                    else:
                        append((Signature,desc_string))
                    append((Signature.Operator, ', '))
                except:
                    #NOTE: This used to be printed...but seeing as this bug never seemed to cause any problems and only spammed the terminal, I'm getting rid of it. I bet you'll never notice...
                    # print("(Just caught a bug that would have crashed the prompt-toolkit gui. Why did this happen??)")
                    pass#I don't know why, but this broke pseudo terminal. I don't know what this is but I don't really care...it's not worth crashing over...

            if sig.params:
                # Pop last comma
                result.pop()

            append((Signature.Operator, ')'))
            append((Signature, ' '))
        return result

    return ConditionalContainer(
        content=Window(
            TokenListControl(get_tokens),
            height=LayoutDimension.exact(1)),
        filter=
            # Show only when there is a signature
            HasSignature(python_input) &
            # And there are no completions to be shown. (would cover signature pop-up.)
            ~(HasCompletions() & (show_completions_menu(python_input) |
                                   show_multi_column_completions_menu(python_input)))
            # Signature needs to be shown.
            & ShowSignature(python_input) &
            # Not done yet.
            ~IsDone())


class PythonPromptMargin(PromptMargin):
    """
    Create margin that displays the prompt.
    It shows something like "In [1]:".
    """
    def __init__(self, python_input):
        self.python_input = python_input

        def get_prompt_style():
            return python_input.all_prompt_styles[python_input.prompt_style]

        def get_prompt(cli):
            return get_prompt_style().in_tokens(cli)

        def get_continuation_prompt(cli, width):
            return get_prompt_style().in2_tokens(cli, width)

        super(PythonPromptMargin, self).__init__(get_prompt, get_continuation_prompt,
                show_numbers=Condition(lambda cli: python_input.show_line_numbers))


def status_bar(python_input):
    """
    Create the `Layout` for the status bar.
    """
    TB = Token.Toolbar.Status

    @if_mousedown
    def toggle_mouse_support(cli, mouse_event):
        python_input.enable_mouse_support = not python_input.enable_mouse_support

    @if_mousedown
    def toggle_microcompletions(cli, mouse_event):
        python_input.enable_microcompletions = not python_input.enable_microcompletions

    @if_mousedown
    def toggle_wrap_lines(cli, mouse_event):
        python_input.wrap_lines = not python_input.wrap_lines

    @if_mousedown
    def toggle_paste_mode(cli, mouse_event):
        python_input.paste_mode = not python_input.paste_mode

    @if_mousedown
    def enter_history(cli, mouse_event):
        python_input.enter_history(cli)

    def get_tokens(cli):
        python_buffer = cli.buffers[DEFAULT_BUFFER]

        result = []
        append = result.append

        if hasattr(python_input,'session_title') and python_input.session_title:
            # append((Token.Window.Title, python_input.session_title, toggle_microcompletions))
            append((TB.PseudoTerminalCurrentVariable, python_input.session_title, toggle_microcompletions))
        else:
            pass
        

        append((TB, ' '))
        result.extend(get_inputmode_tokens(cli, python_input))


        # Position in history.
        append((TB, '%i/%i ' % (python_buffer.working_index + 1,
                                len(python_buffer._working_lines))))

        # Shortcuts.
        if not python_input.vi_mode and cli.current_buffer_name == SEARCH_BUFFER:
            append((TB, '[Ctrl-G] Cancel search [Enter] Go to this position.'))
        elif bool(cli.current_buffer.selection_state) and not python_input.vi_mode:
            # Emacs cut/copy keys.
            append((TB, '[Ctrl-W] Cut [Meta-W] Copy [Ctrl-Y] Paste [Ctrl-G] Cancel'))
        else:


            result.extend([
                (TB.Key, '[F3]', enter_history),
                (TB, 'History ', enter_history),
                (TB.Key, '[F6]', toggle_paste_mode),
                # (TB, ' ', toggle_paste_mode),
            ])

            if python_input.paste_mode:
                append((TB.PasteModeOn, 'Paste:On  ', toggle_paste_mode))
            else:
                append((TB, 'Paste:Off ', toggle_paste_mode))

            result.extend([
                (TB.Key, '[F1]', toggle_mouse_support),
            ])
            if python_input.enable_mouse_support:
                append((TB.PasteModeOn, 'Mouse:On  ', toggle_mouse_support))
            else:
                append((TB, 'Mouse:Off ', toggle_mouse_support))

            

            result.extend([
                (TB.Key, '[F7]', toggle_wrap_lines),
            ])
            if hasattr(python_input,'wrap_lines') and python_input.wrap_lines:
                append((TB, 'Wrap:On  ', toggle_wrap_lines))
            else:
                append((TB.PasteModeOn, 'Wrap:Off ', toggle_wrap_lines))



            result.extend([
                (TB.Key, '[F8]', toggle_microcompletions),
            ])
            if hasattr(python_input,'enable_microcompletions') and python_input.enable_microcompletions:
                append((TB, 'Micro:On  ', toggle_microcompletions))
            else:
                append((TB.PasteModeOn, 'Micro:Off ', toggle_microcompletions))
            

            # if hasattr(python_input,'session_title') and python_input.session_title:
            #     # append((Token.Window.Title, python_input.session_title, toggle_microcompletions))
            #     append((TB.PseudoTerminalCurrentVariable, python_input.session_title, toggle_microcompletions))
            # else:
            #     pass
            



        #region RYAN BURGERT CODE GOES HERE FOR PSEUDOTERMINAL STUFF
        if hasattr(python_input,'show_last_assignable') and python_input.show_last_assignable:
            append((TB, ' '))
            @if_mousedown
            def testytest(cli,mouse_event):
                # python_input.enter_history(cli)
                pass
            import rp.r_iterm_comm
            append((TB.PseudoTerminalCurrentVariable,repr(rp.r_iterm_comm.last_assignable_comm),testytest))
            append((TB, ' '))
        return result

    return TokenListToolbar(
        get_tokens,
        default_char=Char(token=TB),
        filter=~IsDone() & RendererHeightIsKnown() &
            Condition(lambda cli: python_input.show_status_bar and
                                  not python_input.show_exit_confirmation))

selection_flag=False
def get_inputmode_tokens(cli, python_input):
    """
    Return current input mode as a list of (token, text) tuples for use in a
    toolbar.

    :param cli: `CommandLineInterface` instance.
    """
    @if_mousedown
    def toggle_vi_mode(cli, mouse_event):
        python_input.vi_mode = not python_input.vi_mode

    token = Token.Toolbar.Status

    mode = cli.vi_state.input_mode
    result = []
    append = result.append

    append((token.InputMode, '[F4] ', toggle_vi_mode))

    #region  RYAN BURGERT CODE: This code is called every time the user makes a selection. I'm going to use this to chage the selection area to make more sense (in my opinion)
    global selection_flag
    buffer=cli.current_buffer
    has_selection=bool(buffer.selection_state)# Handling on_new_selection...
    if has_selection and not selection_flag:# We made a new selection
        document=buffer.document
        assert isinstance(document,Document)
        assert isinstance(document._selection,SelectionState)
        #
        cursor,cursor_orig=document.cursor_position,document._selection.original_cursor_position
        if cursor>cursor_orig:
            cursor-=1
        elif cursor<cursor_orig:
            cursor_orig-=1
        #
        try:
            buffer.document=Document(document.text,cursor_position=cursor,selection=None)
        except:
            pass
        buffer.document._selection=SelectionState(original_cursor_position=cursor_orig,type=document.selection.type)
        #
        # from rp import text_to_speech
        # text_to_speech("HI")# For debugging
    selection_flag=has_selection
    #endregion

    # InputMode
    if python_input.vi_mode:
        if has_selection:
            if buffer.selection_state.type == SelectionType.LINES:
                append((token.InputMode, 'Vi (VISUAL LINE)', toggle_vi_mode))
            elif buffer.selection_state.type == SelectionType.CHARACTERS:
                append((token.InputMode, 'Vi (VISUAL)', toggle_vi_mode))
                append((token, ' '))
            elif buffer.selection_state.type == 'BLOCK':
                append((token.InputMode, 'Vi (VISUAL BLOCK)', toggle_vi_mode))
                append((token, ' '))
        elif mode in (InputMode.INSERT, 'vi-insert-multiple'):
            append((token.InputMode, 'Vi (INSERT)', toggle_vi_mode))
            append((token, '  '))
        elif mode == InputMode.NAVIGATION:
            append((token.InputMode, 'Vi (NAV)', toggle_vi_mode))
            append((token, '     '))
        elif mode == InputMode.REPLACE:
            append((token.InputMode, 'Vi (REPLACE)', toggle_vi_mode))
            append((token, ' '))
    else:
        append((token.InputMode, 'RP-Emacs', toggle_vi_mode))
        append((token, ' '))

    return result


def show_sidebar_button_info(python_input):
    """
    Create `Layout` for the information in the right-bottom corner.
    (The right part of the status bar.)
    """
    @if_mousedown
    def toggle_sidebar(cli, mouse_event):
        " Click handler for the menu. "
        python_input.show_sidebar = not python_input.show_sidebar

    token = Token.Toolbar.Status

    version = sys.version_info
    import datetime
    remove_beginning_zero=lambda x:x[1:] if x[0] == '0' else x
    import rp

    # if not hasattr(python_input,'session_title'):
    #     title=''
    # else:
    #     title=python_input.session_title
    # TB = Token.Toolbar.Status

    get_tokens=lambda cli: [

        # (Token.Window.Title, title),
        # (token.Key, title),
        # (TB.PseudoTerminalCurrentVariable, title),
        (token, ' '),
        *(
            [
                (token.BatteryPluggedIn if rp.battery_plugged_in() else token.BatteryNotPluggedIn,"🔋 "*0+str(rp.battery_percentage())[:6]+'%'),# put the time here
                (token, ' ')
            ]
            if hasattr(python_input,'show_battery_life') and python_input.show_battery_life
            else []
        ),
        (token,(remove_beginning_zero(datetime.datetime.now().strftime("%I:%M%p")).lower())),# put the time here
        (token, ' '),
        (token.Key, '[F2]', toggle_sidebar),
        (token, ' Menu', toggle_sidebar),
        # (token, ' - '),
        # (token.PythonVersion, '%s %i.%i.%i' % (platform.python_implementation(),
        #                                        version[0], version[1], version[2])),
        (token, ' '),
    ]
    width = token_list_width(get_tokens(None))

    # def get_tokens(cli):
    #     Python version
        # return tokens

    return ConditionalContainer(
        content=Window(
            TokenListControl(get_tokens, default_char=Char(token=token)),
            height=LayoutDimension.exact(1),
            width=LayoutDimension.exact(width)),
        filter=~IsDone() & RendererHeightIsKnown() &
            Condition(lambda cli: python_input.show_status_bar and
                                  not python_input.show_exit_confirmation))


def exit_confirmation(python_input, token=Token.ExitConfirmation):
    """
    Create `Layout` for the exit message.
    """
    def get_tokens(cli):
        # Show "Do you really want to exit?"
        return [
            (token, '\n %s ([y]/n)' % python_input.exit_message),
            (Token.SetCursorPosition, ''),
            (token, '  \n'),
        ]

    visible = ~IsDone() & Condition(lambda cli: python_input.show_exit_confirmation)

    return ConditionalContainer(
        content=Window(TokenListControl(
            get_tokens, default_char=Char(token=token), has_focus=visible)),
        filter=visible)


def meta_enter_message(python_input):
    """
    Create the `Layout` for the 'Meta+Enter` message.
    """
    def get_tokens(cli):
        return [(Token.AcceptMessage, ' [Meta+Enter] Execute ')]

    def extra_condition(cli):
        " Only show when... "
        b = cli.buffers[DEFAULT_BUFFER]

        return (
            python_input.show_meta_enter_message and
            (not b.document.is_cursor_at_the_end or
                python_input.accept_input_on_enter is None) and
            b.is_multiline())

    visible = ~IsDone() & HasFocus(DEFAULT_BUFFER) & Condition(extra_condition)

    return ConditionalContainer(
        content=Window(TokenListControl(get_tokens)),
        filter=visible)


def create_layout(python_input,
                  lexer=PythonLexer,
                  extra_body=None, extra_toolbars=None,
                  extra_buffer_processors=None, input_buffer_height=None):
    D = LayoutDimension
    extra_body = [extra_body] if extra_body else []
    extra_toolbars = extra_toolbars or []
    extra_buffer_processors = extra_buffer_processors or []
    input_buffer_height = input_buffer_height or D(min=6)# This is the space between the input prompt and the toolbar. It lets the autocompletion menus have reaonable height.

    def create_python_input_window():
        def menu_position(cli):
            """
            When there is no autocompletion menu to be shown, and we have a signature,
            set the pop-up position at `bracket_start`.
            """
            b = cli.buffers[DEFAULT_BUFFER]

            if b.complete_state is None and python_input.signatures:
                row, col = python_input.signatures[0].bracket_start
                index = b.document.translate_row_col_to_index(row - 1, col)
                return index

        return Window(
            BufferControl(
                buffer_name=DEFAULT_BUFFER,
                lexer=lexer,
                input_processors=[
                    # Order matters here! We want to ensure brackets are highlighted properly
                    # First highlight selections 
                    HighlightSelectionProcessor(),
                    DisplayMultipleCursors(DEFAULT_BUFFER),
                    # Then highlight search results
                    ConditionalProcessor(
                        processor=HighlightSearchProcessor(preview_search=True),
                        filter=HasFocus(SEARCH_BUFFER),
                    ),
                    # Then highlight matching brackets - this is critical
                    ConditionalProcessor(
                        processor=HighlightMatchingBracketProcessor(chars='[](){}'),
                        filter=HasFocus(DEFAULT_BUFFER) & ~IsDone() &
                            Condition(lambda cli: python_input.highlight_matching_parenthesis)),
                    # Other processors can come after
                    ConditionalProcessor(
                        processor=AppendAutoSuggestion(),
                        filter=~IsDone()),
                    ConditionalProcessor(
                        processor=IndentGuideProcessor(tabstop=4, propagate=False),
                        filter=HasFocus(DEFAULT_BUFFER) & ~IsDone() &
                            Condition(lambda cli: python_input.indent_guides_mode == 'Regular')),
                    ConditionalProcessor(
                        processor=IndentGuideProcessor(tabstop=4, propagate=True),
                        filter=HasFocus(DEFAULT_BUFFER) & ~IsDone() &
                            Condition(lambda cli: python_input.indent_guides_mode == 'Propagate')),
                    ConditionalProcessor(
                        processor=ShowWhitespaceProcessor(token=Token.Whitespace),
                        filter=HasFocus(DEFAULT_BUFFER) & ~IsDone() &
                            Condition(lambda cli: python_input.show_whitespace)),
                    ConditionalProcessor(
                        processor=HighlightWordOccurrencesProcessor(min_word_length=1, skip_keywords=True),
                        filter=HasFocus(DEFAULT_BUFFER) & ~IsDone() &
                            Condition(lambda cli: python_input.highlight_matching_words))
                ] + extra_buffer_processors,
                menu_position=menu_position,

                # Make sure that we always see the result of an reverse-i-search:
                preview_search=True,
            ),
            left_margins=[PythonPromptMargin(python_input)],
            # Scroll offsets. The 1 at the bottom is important to make sure the
            # cursor is never below the "Press [Meta+Enter]" message which is a float.
            scroll_offsets=ScrollOffsets(bottom=1, left=4, right=4),
            # Highlight cursor line - brackets will be drawn on top of this
            # Now using same conditions as other processors like HighlightWordOccurrencesProcessor
            cursorline=Condition(lambda cli: python_input.highlight_cursor_line) & 
                       HasFocus(DEFAULT_BUFFER) & ~IsDone() &
                       Condition(lambda cli: len(cli.buffers[DEFAULT_BUFFER].document.lines) > 1),
            cursorline_token=Token.CursorLine,
            # Highlight cursor column - same conditions as cursor line
            cursorcolumn=Condition(lambda cli: python_input.highlight_cursor_column) & 
                      HasFocus(DEFAULT_BUFFER) & ~IsDone(),
            cursorcolumn_token=Token.CursorColumn,
            # As long as we're editing, prefer a minimal height of 6.
            get_height=(lambda cli: (
                None if cli.is_done or python_input.show_exit_confirmation
                        else input_buffer_height)),
            wrap_lines=Condition(lambda cli: python_input.wrap_lines),
        )
    def title_fill(title,fill=' ',style=Token.Window.TIItleV2):
        height=title.count("\n")+1
        import rp.r_iterm_comm
        rp.r_iterm_comm.python_input_buffers[title]=title
        try:
            return VSplit([
            Window(height=D.exact(height),content=FillControl(fill,token=style)),
            Window(BufferControl(buffer_name=title,lexer=SimpleLexer(default_token=style),),wrap_lines=False,height=D.exact(height),width=D.exact(max(len(x)for x in title.split('\n')))),
            Window(height=D.exact(height),content=FillControl(fill,token=style)),
                    ])
        except:
            return Window(height=D.exact(1),content=FillControl(fill,token=Token.Separator))
    import rp
    return HSplit([
                      VSplit([
                          HSplit([
                              FloatContainer(
                                  content=HSplit(
                                      [create_python_input_window()] + extra_body
                                  ),
                                  floats=[
                                      Float(xcursor=True,
                                            ycursor=True,
                                            content=CompletionsMenu(
                                                scroll_offset=Integer.from_callable(
                                                    lambda:python_input.completion_menu_scroll_offset),
                                                max_height=12,
                                                extra_filter=show_completions_menu(python_input))),
                                      Float(xcursor=True,
                                            ycursor=True,
                                            content=MultiColumnCompletionsMenu(
                                                extra_filter=show_multi_column_completions_menu(python_input))),
                                      Float(xcursor=True,
                                            ycursor=True,
                                            content=signature_toolbar(python_input)),
                                      Float(left=2,
                                            bottom=1,
                                            content=exit_confirmation(python_input)),
                                      Float(bottom=0,right=0,height=1,
                                            content=meta_enter_message(python_input),
                                            hide_when_covering_content=True),
                                      Float(bottom=1,left=1,right=0,content=python_sidebar_help(python_input)),
                                  ]),
                              ArgToolbar(),
                              SearchToolbar(),
                              SystemToolbar(),
                              ValidationToolbar(),
                              CompletionsToolbar(extra_filter=show_completions_toolbar(python_input)),
                              # Docstring region.
                              ConditionalContainer(
                                  content=Window(height=D.exact(1),
                                                 content=FillControl('\u2500',token=Token.Separator)),
                                  filter=HasSignature(python_input) & ShowDocstring(python_input) & ~IsDone()),
                              ConditionalContainer(
                                  content=Window(
                                      BufferControl(
                                          buffer_name='docstring',
                                          lexer=SimpleLexer(default_token=Token.Docstring),
                                          # lexer=PythonLexer,
                                      ),
                                      height=D(max=12)),
                                  filter=HasSignature(python_input) & ShowDocstring(python_input) & ~IsDone(),
                              ),
                              # realtime display region RYAN BURGERT CODE zone
                              # ConditionalContainer(
                              #     content=Window(height=D.exact(1),content=FillControl('\u2500',token=Token.Separator)),
                              #     filter=ShowVarSpaceOrShowRealtimeInput(python_input) & ~IsDone()),
                              ConditionalContainer(


                                  content=
                                  HSplit([
                                      # title_fill("Parenthesis Automator"),
                                      # Window(BufferControl(buffer_name='parenthesizer_buffer',lexer=lexer,),wrap_lines=False,height=D(max=rp.r_iterm_comm.parenthesized_line.count('\n'),min=rp.r_iterm_comm.parenthesized_line.count('\n')    ))
                                      Window(BufferControl(buffer_name='parenthesizer_buffer',lexer=lexer,),wrap_lines=False)
                                      # title_fill(rp.r_iterm_comm.parenthesized_line),
                                  ]),
                                  filter=ShowParenthesisAutomator(python_input) & ~IsDone(),
                              ),
                              VSplit([
                                  ConditionalContainer(
                                      content=
                                      HSplit([
                                          # title_fill("Realtime Evaluator"),
                                          Window(BufferControl(buffer_name='realtime_display',
                                                               # lexer=SimpleLexer(default_token=Token.Docstring)
                                                               lexer=lexer
                                                               ,),wrap_lines=True,height=D(weight=2))
                                      ]),
                                      filter=ShowRealtimeInput(python_input) & ~IsDone(),
                                  ),
                                  ConditionalContainer(
                                      content=Window(width=D.exact(1),
                                                     content=FillControl('│',token=Token.Window.TIItleV2)),
                                      filter=ShowVarSpaceAndShowRealtimeInput(python_input) & ~IsDone()),
                                  ConditionalContainer(


                                      content=
                                      HSplit([
                                          # title_fill("VARS"),
                                          Window(BufferControl(buffer_name='vars',
                                                               lexer=SimpleLexer(default_token=Token.Docstring)
                                                               # lexer=lexer
                                                               ,),wrap_lines=True)
                                      ]),
                                      filter=ShowVarSpace(python_input) & ~IsDone(),
                                  ),
                                ]),
                          ]),
                          HSplit([
                              python_sidebar(python_input),
                              python_sidebar_navigation(python_input),
                          ])
                      ]),
                  ] + extra_toolbars + [
                      VSplit([
                          status_bar(python_input),
                          show_sidebar_button_info(python_input),
                      ])
                  ])



from numbers import Number

from ..object import ObjectNode
from ..array import ArrayNode
from ..primitive import StringNode, IntegerNode, NumericNode, BooleanNode, NullNode, GenericNode

def hasattrs(x,*attrs):
    return all(hasattr(x,attr) for attr in attrs)

def is_numeric(x):
    return isinstance(x,Number) or isinstance(x,type) and issubclass(x,Number)

def is_dictlike(x):
    return hasattrs(x,'__getitem__','__iter__','__len__','keys')

def is_arraylike(x):
    return hasattrs(x,'__getitem__','__iter__','__len__')


class DefaultImplementationSelector:
    def __getitem__(self,key):
        if key in DEFAULT_JSON_NODE_IMPLS:
            return DEFAULT_JSON_NODE_IMPLS[key]
        if is_numeric(key):
            return NumericNode
        if is_dictlike(key):
            return ObjectNode
        if is_arraylike(key):
            return ArrayNode
        return GenericNode


DEFAULT_JSON_NODE_IMPLS = {
    list: ArrayNode,
    dict: ObjectNode,
    str: StringNode,
    int: IntegerNode,
    bool: BooleanNode,
    float: NumericNode,
    type(None): NullNode
}



from typing import Any

class HTTPError(Exception): ...
class HTTPWarning(Warning): ...

class PoolError(HTTPError):
    pool: Any
    def __init__(self, pool, message) -> None: ...
    def __reduce__(self): ...

class RequestError(PoolError):
    url: Any
    def __init__(self, pool, url, message) -> None: ...
    def __reduce__(self): ...

class SSLError(HTTPError): ...
class ProxyError(HTTPError): ...
class DecodeError(HTTPError): ...
class ProtocolError(HTTPError): ...

ConnectionError: Any

class MaxRetryError(RequestError):
    reason: Any
    def __init__(self, pool, url, reason=...) -> None: ...

class HostChangedError(RequestError):
    retries: Any
    def __init__(self, pool, url, retries=...) -> None: ...

class TimeoutStateError(HTTPError): ...
class TimeoutError(HTTPError): ...
class ReadTimeoutError(TimeoutError, RequestError): ...
class ConnectTimeoutError(TimeoutError): ...
class EmptyPoolError(PoolError): ...
class ClosedPoolError(PoolError): ...
class LocationValueError(ValueError, HTTPError): ...

class LocationParseError(LocationValueError):
    location: Any
    def __init__(self, location) -> None: ...

class ResponseError(HTTPError):
    GENERIC_ERROR: Any
    SPECIFIC_ERROR: Any

class SecurityWarning(HTTPWarning): ...
class InsecureRequestWarning(SecurityWarning): ...
class SystemTimeWarning(SecurityWarning): ...
class InsecurePlatformWarning(SecurityWarning): ...
